from abc import ABCMeta
from typing import Optional, Type, Tuple

import torch
from torch import Tensor
from torch import nn, autograd
from torch.distributions.utils import lazy_property
from torch.nn import init
from torch.nn.utils.rnn import PackedSequence
from torchrua import TreeReduceIndices, tree_reduce_packed_indices
from torchrua import select_head, select_last, roll_packed_sequence, pad_packed_sequence

from torchlatent.semiring import Semiring, Log, Max

__all__ = [
    'compute_scores',
    'compute_partitions',
    'CrfDistribution',
    'CrfDecoderABC', 'CrfDecoder',
]


def compute_scores(semiring: Type[Semiring]):
    def _compute_scores(
            emissions: PackedSequence, tags: PackedSequence,
            transitions: Tensor, head_transitions: Tensor, tail_transitions: Tensor) -> Tensor:
        device = transitions.device

        emission_scores = emissions.data.gather(dim=-1, index=tags.data[..., None])[..., 0]  # [t, c]

        h = emissions.batch_sizes[0].item()
        t = torch.arange(transitions.size()[0], device=device)  # [t]
        c = torch.arange(transitions.size()[1], device=device)  # [c]

        x, y = roll_packed_sequence(tags, shifts=1).data, tags.data  # [t, c]
        head = select_head(tags, unsort=False)  # [h, c]
        tail = select_last(tags, unsort=False)  # [h, c]

        transition_scores = transitions[t[:, None], c[None, :], x, y]  # [t, c]
        transition_head_scores = head_transitions[t[:h, None], c[None, :], head]  # [h, c]
        transition_tail_scores = tail_transitions[t[:h, None], c[None, :], tail]  # [h, c]

        transition_scores[:h] = transition_head_scores  # [h, c]
        scores, _ = pad_packed_sequence(
            PackedSequence(
                data=semiring.mul(emission_scores, transition_scores),
                batch_sizes=emissions.batch_sizes,
                sorted_indices=None,
                unsorted_indices=None,
            ),
            batch_first=False,
        )
        scores = semiring.prod(scores, dim=0)
        scores = semiring.mul(scores, transition_tail_scores)

        if emissions.unsorted_indices is not None:
            scores = scores[emissions.unsorted_indices]

        return scores

    return _compute_scores


def compute_partitions(semiring: Type[Semiring]):
    def _compute_partitions(
            emissions: PackedSequence, indices: TreeReduceIndices,
            transitions: Tensor, head_transitions: Tensor, tail_transitions: Tensor, eye: Tensor) -> Tensor:
        h = emissions.batch_sizes[0].item()
        t = torch.arange(transitions.size()[0], device=transitions.device)  # [t]
        c = torch.arange(transitions.size()[1], device=transitions.device)  # [c]

        scores = semiring.mul(transitions, emissions.data[..., None, :])  # [t, c, n, n]
        scores[:h] = eye[None, None, :, :]
        scores = semiring.reduce(tensor=scores, indices=indices)

        emission_head_scores = emissions.data[:h, :, None, :]
        transition_head_scores = head_transitions[t[:h, None], c[None, :], None, :]
        transition_tail_scores = tail_transitions[t[:h, None], c[None, :], :, None]

        scores = semiring.bmm(
            semiring.bmm(semiring.mul(transition_head_scores, emission_head_scores), scores),
            transition_tail_scores,
        )[..., 0, 0]

        if emissions.unsorted_indices is not None:
            scores = scores[emissions.unsorted_indices]
        return scores

    return _compute_partitions


class CrfDistribution(object):
    def __init__(self, emissions: PackedSequence, indices: TreeReduceIndices,
                 transitions: Tensor, head_transitions: Tensor, tail_transitions: Tensor) -> None:
        super(CrfDistribution, self).__init__()
        self.emissions = emissions
        self.indices = indices

        self.transitions = transitions
        self.head_transitions = head_transitions
        self.tail_transitions = tail_transitions

    def semiring_scores(self, semiring: Type[Semiring], tags: PackedSequence) -> Tensor:
        return compute_scores(semiring=semiring)(
            emissions=self.emissions, tags=tags,
            transitions=self.transitions,
            head_transitions=self.head_transitions,
            tail_transitions=self.tail_transitions,
        )

    def semiring_partitions(self, semiring: Type[Semiring]) -> Tensor:
        return compute_partitions(semiring=semiring)(
            emissions=self.emissions, indices=self.indices,
            transitions=self.transitions,
            head_transitions=self.head_transitions,
            tail_transitions=self.tail_transitions,
            eye=semiring.eye_like(self.transitions),
        )

    def log_prob(self, tags: PackedSequence) -> Tensor:
        return self.log_scores(tags=tags) - self.log_partitions

    def log_scores(self, tags: PackedSequence) -> Tensor:
        return self.semiring_scores(semiring=Log, tags=tags)

    @lazy_property
    def log_partitions(self) -> Tensor:
        return self.semiring_partitions(semiring=Log)

    @lazy_property
    def marginals(self) -> Tensor:
        log_partitions = self.log_partitions
        grad, = autograd.grad(
            log_partitions, self.emissions.data, torch.ones_like(log_partitions),
            create_graph=True, only_inputs=True, allow_unused=False,
        )
        return grad

    @lazy_property
    def argmax(self) -> PackedSequence:
        max_partitions = self.semiring_partitions(semiring=Max)

        grad, = torch.autograd.grad(
            max_partitions, self.emissions.data, torch.ones_like(max_partitions),
            retain_graph=False, create_graph=False, allow_unused=False,
        )
        return PackedSequence(
            data=grad.argmax(dim=-1),
            batch_sizes=self.emissions.batch_sizes,
            sorted_indices=self.emissions.sorted_indices,
            unsorted_indices=self.emissions.unsorted_indices,
        )


class CrfDecoderABC(nn.Module, metaclass=ABCMeta):
    def __init__(self, num_tags: int, num_conjugates: int):
        super(CrfDecoderABC, self).__init__()

        self.num_tags = num_tags
        self.num_conjugates = num_conjugates

    def reset_parameters(self) -> None:
        raise NotImplementedError

    def extra_repr(self) -> str:
        return ', '.join([
            f'num_tags={self.num_tags}',
            f'num_conjugates={self.num_conjugates}',
        ])

    def compile_indices(self, emissions: PackedSequence, tags: Optional[PackedSequence] = None,
                        indices: Optional[TreeReduceIndices] = None, **kwargs):
        assert emissions.data.dim() == 3, f'{emissions.data.dim()} != {3}'
        if tags is not None:
            assert tags.data.dim() == 2, f'{tags.data.dim()} != {2}'

        if indices is None:
            batch_sizes = emissions.batch_sizes.to(device=emissions.data.device)
            indices = tree_reduce_packed_indices(batch_sizes=batch_sizes)

        return indices

    def obtain_parameters(self, **kwargs) -> Tuple[Tensor, Tensor, Tensor]:
        return self.transitions, self.head_transitions, self.tail_transitions

    def forward(self, emissions: PackedSequence, tags: Optional[PackedSequence] = None,
                indices: Optional[TreeReduceIndices] = None, **kwargs):
        indices = self.compile_indices(emissions=emissions, tags=tags, indices=indices)
        transitions, head_transitions, tail_transitions = self.obtain_parameters(
            emissions=emissions, tags=tags, indices=indices,
        )

        dist = CrfDistribution(
            emissions=emissions, indices=indices,
            transitions=transitions,
            head_transitions=head_transitions,
            tail_transitions=tail_transitions,
        )

        return dist, tags

    def fit(self, emissions: PackedSequence, tags: PackedSequence,
            indices: Optional[TreeReduceIndices] = None, **kwargs) -> Tensor:
        dist, tags = self(emissions=emissions, tags=tags, instr=indices, **kwargs)

        return dist.log_prob(tags=tags)

    def decode(self, emissions: PackedSequence,
               indices: Optional[TreeReduceIndices] = None, **kwargs) -> PackedSequence:
        dist, _ = self(emissions=emissions, tags=None, instr=indices, **kwargs)
        return dist.argmax

    def marginals(self, emissions: PackedSequence,
                  indices: Optional[TreeReduceIndices] = None, **kwargs) -> Tensor:
        dist, _ = self(emissions=emissions, tags=None, instr=indices, **kwargs)
        return dist.marginals


class CrfDecoder(CrfDecoderABC):
    def __init__(self, num_tags: int, num_conjugates: int = 1) -> None:
        super(CrfDecoder, self).__init__(num_tags=num_tags, num_conjugates=num_conjugates)

        self.transitions = nn.Parameter(
            torch.empty((1, self.num_conjugates, self.num_tags, self.num_tags)),
            requires_grad=True,
        )
        self.head_transitions = nn.Parameter(
            torch.empty((1, self.num_conjugates, self.num_tags)),
            requires_grad=True,
        )
        self.tail_transitions = nn.Parameter(
            torch.empty((1, self.num_conjugates, self.num_tags)),
            requires_grad=True,
        )

        self.reset_parameters()

    @torch.no_grad()
    def reset_parameters(self) -> None:
        bound = 0.01
        init.uniform_(self.transitions, -bound, +bound)
        init.uniform_(self.head_transitions, -bound, +bound)
        init.uniform_(self.tail_transitions, -bound, +bound)



from typing import Optional

import torch
from torch import Tensor
from torch.nn.utils.rnn import PackedSequence


def broadcast_packed_sequences(
        emissions: PackedSequence, tags: Optional[PackedSequence],
        transitions: Tensor, head_transitions: Tensor, tail_transitions: Tensor):
    """
        Args:
            emissions: [t1, c1, n]
            tags: [t1, c1]
            transitions: [t2, c2, n, n]
            head_transitions: [h2, c2, n]
            tail_transitions: [h2, c2, n]
    """
    assert emissions.data.dim() == 3, f'{emissions.data.size()}'
    assert transitions.dim() == 4, f'{transitions.size()}'
    assert head_transitions.dim() == 3, f'{head_transitions.size()}'
    assert tail_transitions.dim() == 3, f'{tail_transitions.size()}'

    _, _, n = emissions.data.size()
    h = emissions.batch_sizes[0].item()

    if tags is None:
        t, c, = torch.broadcast_shapes(
            emissions.data.size()[:2],
            transitions.size()[:2],
        )
    else:
        assert tags.data.dim() == 2, f'{tags.data.size()}'

        t, c, = torch.broadcast_shapes(
            emissions.data.size()[:2],
            tags.data.size()[:2],
            transitions.size()[:2],
        )
        tags = tags._replace(data=tags.data.expand((t, c)))

    emissions = emissions._replace(data=emissions.data.expand((t, c, n)))
    transitions = transitions.expand((t, c, n, n))
    head_transitions = head_transitions.expand((h, c, n))
    tail_transitions = tail_transitions.expand((h, c, n))

    return emissions, tags, transitions, head_transitions, tail_transitions, (t, c, n, h)



import torch
from torch import Tensor
from torchrua.tree_reduction import tree_reduce_sequence, TreeReduceIndices

from torchlatent.functional import logsumexp

__all__ = [
    'Semiring',
    'Std', 'Log', 'Max',
]


class Semiring(object):
    zero: float
    one: float

    @classmethod
    def eye_like(cls, tensor: Tensor) -> Tensor:
        *_, n = tensor.size()
        eye = torch.full((n, n), fill_value=cls.zero, dtype=tensor.dtype, device=tensor.device)
        index = torch.arange(n, dtype=torch.long, device=tensor.device)
        eye[index, index] = cls.one
        return eye

    @classmethod
    def add(cls, x: Tensor, y: Tensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def mul(cls, x: Tensor, y: Tensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def sum(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        raise NotImplementedError

    @classmethod
    def prod(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        raise NotImplementedError

    @classmethod
    def bmm(cls, x: Tensor, y: Tensor) -> Tensor:
        return cls.sum(cls.mul(x.unsqueeze(-1), y.unsqueeze(-3)), dim=-2, keepdim=False)

    @classmethod
    def reduce(cls, tensor: Tensor, indices: TreeReduceIndices) -> Tensor:
        return tree_reduce_sequence(cls.bmm)(tensor=tensor, indices=indices)


class Std(Semiring):
    zero = 0.
    one = 1.

    @classmethod
    def add(cls, x: Tensor, y: Tensor) -> Tensor:
        return x + y

    @classmethod
    def mul(cls, x: Tensor, y: Tensor) -> Tensor:
        return x * y

    @classmethod
    def sum(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.sum(tensor, dim=dim, keepdim=keepdim)

    @classmethod
    def prod(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.prod(tensor, dim=dim, keepdim=keepdim)


class Log(Semiring):
    zero = -float('inf')
    one = 0.

    @classmethod
    def add(cls, x: Tensor, y: Tensor) -> Tensor:
        with torch.no_grad():
            m = torch.maximum(x, y)
            m = m.masked_fill_(torch.isneginf(m), 0.)

        z = (x - m).exp_() + (y - m).exp_()
        mask = z == 0
        z = z.masked_fill_(mask, 1.).log_()
        return z.masked_fill_(mask, -float('inf')).add_(m)

    @classmethod
    def mul(cls, x: Tensor, y: Tensor) -> Tensor:
        return x + y

    @classmethod
    def sum(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return logsumexp(tensor, dim=dim, keepdim=keepdim)

    @classmethod
    def prod(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.sum(tensor, dim=dim, keepdim=keepdim)


class Max(Semiring):
    zero = -float('inf')
    one = 0.

    @classmethod
    def add(cls, x: Tensor, y: Tensor) -> Tensor:
        return torch.maximum(x, y)

    @classmethod
    def mul(cls, x: Tensor, y: Tensor) -> Tensor:
        return x + y

    @classmethod
    def sum(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.max(tensor, dim=dim, keepdim=keepdim)[0]

    @classmethod
    def prod(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.sum(tensor, dim=dim, keepdim=keepdim)



from setuptools import setup, find_packages

name = 'torchlatent'

setup(
    name=name,
    version='0.4.0',
    packages=[package for package in find_packages() if package.startswith(name)],
    url='https://github.com/speedcell4/torchlatent',
    license='MIT',
    author='speedcell4',
    author_email='speedcell4@gmail.com',
    description='High Performance Structured Prediction in PyTorch',
    python_requires='>=3.7',
    install_requires=[
        'numpy',
        'torchrua>=0.3.0',
    ],
    extras_require={
        'dev': [
            'einops',
            'pytest',
            'hypothesis',
            'pytorch-crf',
        ],
    }
)



import torch
from torch import Tensor

__all__ = [
    'logsumexp',
]


def logsumexp(tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
    with torch.no_grad():
        m, _ = torch.max(tensor, dim=dim, keepdim=True)
        m = m.masked_fill_(torch.isneginf(m), 0.)

    z = (tensor - m).exp_().sum(dim=dim, keepdim=True)
    mask = z == 0
    z = z.masked_fill_(mask, 1.).log_()
    z = z.masked_fill_(mask, -float('inf')).add_(m)

    if not keepdim:
        z = z.squeeze(dim=dim)
    return z
